(*
	Copyright (c) 2000
		Cambridge University Technical Services Limited

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.
	
	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)


signature XCURSORS_SIG = sig

  val XC_num_glyphs:          int ;
  val XC_X_cursor:            int ;
  val XC_arrow:               int ;
  val XC_based_arrow_down:    int ;
  val XC_based_arrow_up:      int ;
  val XC_boat:                int ;
  val XC_bogosity:            int ;
  val XC_bottom_left_corner:  int ;
  val XC_bottom_right_corner: int ;
  val XC_bottom_side:         int ;
  val XC_bottom_tee:          int ;
  val XC_box_spiral:          int ;
  val XC_center_ptr:          int ;
  val XC_circle:              int ;
  val XC_clock:               int ;
  val XC_coffee_mug:          int ;
  val XC_cross:               int ;
  val XC_cross_reverse:       int ;
  val XC_crosshair:           int ;
  val XC_diamond_cross:       int ;
  val XC_dot:                 int ;
  val XC_dotbox:              int ;
  val XC_double_arrow:        int ;
  val XC_draft_large:         int ;
  val XC_draft_small:         int ;
  val XC_draped_box:          int ;
  val XC_exchange:            int ;
  val XC_fleur:               int ;
  val XC_gobbler:             int ;
  val XC_gumby:               int ;
  val XC_hand1:               int ;
  val XC_hand2:               int ;
  val XC_heart:               int ;
  val XC_icon:                int ;
  val XC_iron_cross:          int ;
  val XC_left_ptr:            int ;
  val XC_left_side:           int ;
  val XC_left_tee:            int ;
  val XC_leftbutton:          int ;
  val XC_ll_angle:            int ;
  val XC_lr_angle:            int ;
  val XC_man:                 int ;
  val XC_middlebutton:        int ;
  val XC_mouse:               int ;
  val XC_pencil:              int ;
  val XC_pirate:              int ;
  val XC_plus:                int ;
  val XC_question_arrow:      int ;
  val XC_right_ptr:           int ;
  val XC_right_side:          int ;
  val XC_right_tee:           int ;
  val XC_rightbutton:         int ;
  val XC_rtl_logo:            int ;
  val XC_sailboat:            int ;
  val XC_sb_down_arrow:       int ;
  val XC_sb_h_double_arrow:   int ;
  val XC_sb_left_arrow:       int ;
  val XC_sb_right_arrow:      int ;
  val XC_sb_up_arrow:         int ;
  val XC_sb_v_double_arrow:   int ;
  val XC_shuttle:             int ;
  val XC_sizing:              int ;
  val XC_spider:              int ;
  val XC_spraycan:            int ;
  val XC_star:                int ;
  val XC_target:              int ;
  val XC_tcross:              int ;
  val XC_top_left_arrow:      int ;
  val XC_top_left_corner:     int ;
  val XC_top_right_corner:    int ;
  val XC_top_side:            int ;
  val XC_top_tee:             int ;
  val XC_trek:                int ;
  val XC_ul_angle:            int ;
  val XC_umbrella:            int ;
  val XC_ur_angle:            int ;
  val XC_watch:               int ;
  val XC_xterm:               int ;

end (* XCURSORS_SIG *) ;
