(*
	Copyright (c) 2000
		Cambridge University Technical Services Limited

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.
	
	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)


structure XCursors: XCURSORS_SIG = struct

  val XC_num_glyphs           = 154 ;
  val XC_X_cursor             = 0 ;
  val XC_arrow                = 2 ;
  val XC_based_arrow_down     = 4 ;
  val XC_based_arrow_up       = 6 ;
  val XC_boat                 = 8 ;
  val XC_bogosity             = 10 ;
  val XC_bottom_left_corner   = 12 ;
  val XC_bottom_right_corner  = 14 ;
  val XC_bottom_side          = 16 ;
  val XC_bottom_tee           = 18 ;
  val XC_box_spiral           = 20 ;
  val XC_center_ptr           = 22 ;
  val XC_circle               = 24 ;
  val XC_clock                = 26 ;
  val XC_coffee_mug           = 28 ;
  val XC_cross                = 30 ;
  val XC_cross_reverse        = 32 ;
  val XC_crosshair            = 34 ;
  val XC_diamond_cross        = 36 ;
  val XC_dot                  = 38 ;
  val XC_dotbox               = 40 ;
  val XC_double_arrow         = 42 ;
  val XC_draft_large          = 44 ;
  val XC_draft_small          = 46 ;
  val XC_draped_box           = 48 ;
  val XC_exchange             = 50 ;
  val XC_fleur                = 52 ;
  val XC_gobbler              = 54 ;
  val XC_gumby                = 56 ;
  val XC_hand1                = 58 ;
  val XC_hand2                = 60 ;
  val XC_heart                = 62 ;
  val XC_icon                 = 64 ;
  val XC_iron_cross           = 66 ;
  val XC_left_ptr             = 68 ;
  val XC_left_side            = 70 ;
  val XC_left_tee             = 72 ;
  val XC_leftbutton           = 74 ;
  val XC_ll_angle             = 76 ;
  val XC_lr_angle             = 78 ;
  val XC_man                  = 80 ;
  val XC_middlebutton         = 82 ;
  val XC_mouse                = 84 ;
  val XC_pencil               = 86 ;
  val XC_pirate               = 88 ;
  val XC_plus                 = 90 ;
  val XC_question_arrow       = 92 ;
  val XC_right_ptr            = 94 ;
  val XC_right_side           = 96 ;
  val XC_right_tee            = 98 ;
  val XC_rightbutton          = 100 ;
  val XC_rtl_logo             = 102 ;
  val XC_sailboat             = 104 ;
  val XC_sb_down_arrow        = 106 ;
  val XC_sb_h_double_arrow    = 108 ;
  val XC_sb_left_arrow        = 110 ;
  val XC_sb_right_arrow       = 112 ;
  val XC_sb_up_arrow          = 114 ;
  val XC_sb_v_double_arrow    = 116 ;
  val XC_shuttle              = 118 ;
  val XC_sizing               = 120 ;
  val XC_spider               = 122 ;
  val XC_spraycan             = 124 ;
  val XC_star                 = 126 ;
  val XC_target               = 128 ;
  val XC_tcross               = 130 ;
  val XC_top_left_arrow       = 132 ;
  val XC_top_left_corner      = 134 ;
  val XC_top_right_corner     = 136 ;
  val XC_top_side             = 138 ;
  val XC_top_tee              = 140 ;
  val XC_trek                 = 142 ;
  val XC_ul_angle             = 144 ;
  val XC_umbrella             = 146 ;
  val XC_ur_angle             = 148 ;
  val XC_watch                = 150 ;
  val XC_xterm                = 152 ;

end (* XCursors *) ;
