(*
	Copyright (c) 2000
		Cambridge University Technical Services Limited

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.
	
	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.
	
	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(*
    Title: 	Make the ML system.
    Author: 	Mike Crawley
    Copyright   AHL 1993
*)

structure CompilerBody = 
  COMPILER_BODY (
    structure SYMBOLS    = Symbols
    structure SYMSET     = SymSet
    structure LEX        = Lex
    structure CODETREE   = CodeTree
    structure STRUCTVALS = StructVals
    structure VALUEOPS   = ValueOps
    structure TYPETREE   = TypeTree
    structure STRUCTURES = Structures
    structure PARSEDEC   = ParseDec
    structure DEBUG      = Debug
    structure UTILITIES  = Utilities
    
    structure MISC       = Misc
    structure PRETTYPRINTER = PrettyPrinter
    structure POLYML     = PolyML
    structure ADDRESS    = Address
    structure RUNCALL    = RunCall
    structure CREATEDATABASE = CreateDatabase
) ;

